"""Classes for linear optimization constraints.

Linear objective functions must be of the form `A*x-b`, where:
    - `A` is a constant matrix that can be pre-computed
    - `x` is an array of a single parameter
    - `b` is the desired vector set by `objective.target`
"""

import warnings

import numpy as np
from termcolor import colored

from desc.backend import execute_on_cpu, jnp, tree_leaves, tree_map, tree_structure
from desc.basis import zernike_radial
from desc.geometry import FourierRZCurve
from desc.utils import broadcast_tree, errorif, setdefault

from .normalization import compute_scaling_factors
from .objective_funs import _Objective


# TODO (#1391): get rid of this class and inherit from FixParameters instead?
class _FixedObjective(_Objective):
    _fixed = True
    _linear = True
    _scalar = False

    def update_target(self, thing):
        """Update target values using an Optimizable object.

        Parameters
        ----------
        thing : Optimizable
            Optimizable object that will be optimized to satisfy the Objective.

        """
        new_target = self.compute(thing.params_dict)
        assert len(new_target) == len(self.target)
        self.target = new_target
        self._target_from_user = self.target  # in case the Objective is re-built
        if not self._use_jit:
            self._unjit()

    def _parse_target_from_user(
        self, target_from_user, default_target, default_bounds, idx
    ):
        # TODO (#1391): add logic here to deal with `target_from_user` as a pytree?
        # TODO (#1391: does this actually need idx?
        if target_from_user is None:
            target = default_target
            bounds = default_bounds
        elif isinstance(target_from_user, tuple) and (
            len(target_from_user) == 2
        ):  # treat as bounds
            target = None
            bounds = (
                np.broadcast_to(target_from_user[0], self._dim_f).copy()[idx],
                np.broadcast_to(target_from_user[1], self._dim_f).copy()[idx],
            )
        else:
            target = np.broadcast_to(target_from_user, self._dim_f).copy()[idx]
            bounds = None
        return target, bounds


class FixParameters(_Objective):
    """Fix specific degrees of freedom associated with a given Optimizable thing.

    Parameters
    ----------
    thing : Optimizable
        Object whose degrees of freedom are being fixed.
    params : nested list of dicts
        Dict keys are the names of parameters to fix (str), and dict values are the
        indices to fix for each corresponding parameter (int array).
        Use True (False) instead of an int array to fix all (none) of the indices
        for that parameter.
        Must have the same pytree structure as thing.params_dict.
        The default is to fix all indices of all parameters.
    target : dict of {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Should have the same tree structure as thing.params. Defaults to things.params.
    bounds : tuple of dict {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Should have the same tree structure as thing.params.
    weight : dict of {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Should be a scalar or have the same tree structure as thing.params.
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True. Has no effect for this objective.
    name : str, optional
        Name of the objective function.

    Examples
    --------
    .. code-block:: python

        import numpy as np
        from desc.coils import (
            CoilSet, FourierPlanarCoil, FourierRZCoil, FourierXYZCoil, MixedCoilSet
        )
        from desc.objectives import FixParameters

        # toroidal field coil set with 4 coils
        tf_coil = FourierPlanarCoil(
            current=3, center=[2, 0, 0], normal=[0, 1, 0], r_n=[1]
        )
        tf_coilset = CoilSet.linspaced_angular(tf_coil, n=4)
        # vertical field coil set with 3 coils
        vf_coil = FourierRZCoil(current=-1, R_n=3, Z_n=-1)
        vf_coilset = CoilSet.linspaced_linear(
            vf_coil, displacement=[0, 0, 2], n=3, endpoint=True
        )
        # another single coil
        xyz_coil = FourierXYZCoil(current=2)
        # full coil set with TF coils, VF coils, and other single coil
        full_coilset = MixedCoilSet((tf_coilset, vf_coilset, xyz_coil))

        params = [
            [
                {"current": True},  # fix "current" of the 1st TF coil
                # fix "center" and one component of "normal" for the 2nd TF coil
                {"center": True, "normal": np.array([1])},
                {"r_n": True},  # fix radius of the 3rd TF coil
                {},  # fix nothing in the 4th TF coil
            ],
            {"shift": True, "rotmat": True},  # fix "shift" & "rotmat" for all VF coils
            # fix specified indices of "X_n" and "Z_n", but not "Y_n", for other coil
            {"X_n": np.array([1, 2]), "Y_n": False, "Z_n": np.array([0])},
        ]
        obj = FixParameters(full_coilset, params)

    """

    _scalar = False
    _linear = True
    _fixed = True
    _units = "(~)"
    _print_value_fmt = "Fixed parameters error: "

    def __init__(
        self,
        thing,
        params=None,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed parameters",
    ):
        self._params = params
        super().__init__(
            things=thing,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        thing = self.things[0]

        # default params
        default_params = tree_map(lambda dim: np.arange(dim), thing.dimensions)
        self._params = setdefault(self._params, default_params)
        self._params = broadcast_tree(self._params, default_params)
        self._indices = tree_leaves(self._params)
        assert tree_structure(self._params) == tree_structure(default_params)

        self._dim_f = sum(idx.size for idx in self._indices)

        # default target
        if self.target is None and self.bounds is None:
            self.target = np.concatenate(
                [
                    np.atleast_1d(param[idx])
                    for param, idx in zip(tree_leaves(thing.params_dict), self._indices)
                ]
            )

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed degree of freedom errors.

        Parameters
        ----------
        params : list of dict
            List of dictionaries of degrees of freedom, eg CoilSet.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed degree of freedom errors.

        """
        return jnp.concatenate(
            [
                jnp.atleast_1d(param[idx])
                for param, idx in zip(tree_leaves(params), self._indices)
            ]
        )

    def update_target(self, thing):
        """Update target values using an Optimizable object.

        Parameters
        ----------
        thing : Optimizable
            Optimizable object that will be optimized to satisfy the Objective.

        """
        self.target = self.compute(thing.params_dict)
        if not self._use_jit:
            self._unjit()


class BoundaryRSelfConsistency(_Objective):
    """Ensure that the boundary and interior surfaces are self-consistent.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    surface_label : float, optional
        Surface to enforce boundary conditions on. Defaults to Equilibrium.surface.rho
    name : str, optional
        Name of the objective function.

    """

    _scalar = False
    _linear = True
    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "R boundary self consistency error: "

    def __init__(
        self,
        eq,
        surface_label=None,
        name="self_consistency R",
    ):
        self._surface_label = surface_label
        super().__init__(
            things=eq,
            target=0,
            bounds=None,
            weight=1,
            normalize=False,
            normalize_target=False,
            name=name,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        modes = eq.surface.R_basis.modes
        idx = np.arange(eq.surface.R_basis.num_modes)

        self._dim_f = idx.size
        self._A = np.zeros((self._dim_f, eq.R_basis.num_modes))
        Js = []
        surf = eq.surface.rho if self._surface_label is None else self._surface_label
        for i, (l, m, n) in enumerate(eq.R_basis.modes):
            if eq.bdry_mode == "lcfs":
                j = np.argwhere((modes[:, 1:] == [m, n]).all(axis=1))
                Js.append(j.flatten())
            else:
                raise NotImplementedError(
                    "bdry_mode is not lcfs, yell at Dario to finish poincare stuff"
                )
        Js = np.array(Js)
        # Broadcasting at once is faster. We need to use np.arange to avoid
        # setting the value to the whole row.
        self._A[Js[:, 0], np.arange(eq.R_basis.num_modes)] = zernike_radial(
            surf, eq.R_basis.modes[:, 0], eq.R_basis.modes[:, 1]
        )
        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute boundary R self-consistency errors.

        IE, the mismatch between the Fourier-Zernike basis evaluated at rho=1 and the
        double Fourier series defining the equilibrium LCFS

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            boundary R self-consistency errors.

        """
        return jnp.dot(self._A, params["R_lmn"]) - params["Rb_lmn"]


class BoundaryZSelfConsistency(_Objective):
    """Ensure that the boundary and interior surfaces are self consistent.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    surface_label : float, optional
        Surface to enforce boundary conditions on. Defaults to Equilibrium.surface.rho
    name : str, optional
        Name of the objective function.

    """

    _scalar = False
    _linear = True
    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "Z boundary self consistency error: "

    def __init__(
        self,
        eq,
        surface_label=None,
        name="self_consistency Z",
    ):
        self._surface_label = surface_label
        super().__init__(
            things=eq,
            target=0,
            bounds=None,
            weight=1,
            normalize=False,
            normalize_target=False,
            name=name,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        eq : Equilibrium, optional
            Equilibrium that will be optimized to satisfy the Objective.
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        modes = eq.surface.Z_basis.modes
        idx = np.arange(eq.surface.Z_basis.num_modes)

        self._dim_f = idx.size
        self._A = np.zeros((self._dim_f, eq.Z_basis.num_modes))
        Js = []
        surf = eq.surface.rho if self._surface_label is None else self._surface_label
        for i, (l, m, n) in enumerate(eq.Z_basis.modes):
            if eq.bdry_mode == "lcfs":
                j = np.argwhere((modes[:, 1:] == [m, n]).all(axis=1))
                Js.append(j.flatten())
            else:
                raise NotImplementedError(
                    "bdry_mode is not lcfs, yell at Dario to finish poincare stuff"
                )
        Js = np.array(Js)
        # Broadcasting at once is faster. We need to use np.arange to avoid
        # setting the value to the whole row.
        self._A[Js[:, 0], np.arange(eq.Z_basis.num_modes)] = zernike_radial(
            surf, eq.Z_basis.modes[:, 0], eq.Z_basis.modes[:, 1]
        )
        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute boundary Z self-consistency errors.

        IE, the mismatch between the Fourier-Zernike basis evaluated at rho=1 and the
        double Fourier series defining the equilibrium LCFS

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            boundary Z self-consistency errors.

        """
        return jnp.dot(self._A, params["Z_lmn"]) - params["Zb_lmn"]


class AxisRSelfConsistency(_Objective):
    """Ensure consistency between Zernike and Fourier coefficients on axis.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    name : str, optional
        Name of the objective function.

    """

    _scalar = False
    _linear = True
    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "R axis self consistency error: "

    def __init__(
        self,
        eq,
        name="axis R self consistency",
    ):
        super().__init__(
            things=eq,
            target=0,
            weight=1,
            name=name,
            normalize=False,
            normalize_target=False,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        ns = eq.axis.R_basis.modes[:, 2]
        self._dim_f = ns.size
        self._A = np.zeros((self._dim_f, eq.R_basis.num_modes))

        for i, (l, m, n) in enumerate(eq.R_basis.modes):
            if m != 0:
                continue
            if (l // 2) % 2 == 0:
                j = np.argwhere(n == ns)
                self._A[j, i] = 1
            else:
                j = np.argwhere(n == ns)
                self._A[j, i] = -1

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute axis R self-consistency errors.

        IE, the mismatch between the Fourier-Zernike basis evaluated at rho=0 and the
        Fourier series defining the equilibrium axis position

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            axis R self-consistency errors.

        """
        f = jnp.dot(self._A, params["R_lmn"]) - params["Ra_n"]
        return f


class AxisZSelfConsistency(_Objective):
    """Ensure consistency between Zernike and Fourier coefficients on axis.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    name : str, optional
        Name of the objective function.

    """

    _scalar = False
    _linear = True
    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "Z axis self consistency error: "

    def __init__(
        self,
        eq,
        name="axis Z self consistency",
    ):
        super().__init__(
            things=eq,
            target=0,
            weight=1,
            name=name,
            normalize=False,
            normalize_target=False,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        ns = eq.axis.Z_basis.modes[:, 2]
        self._dim_f = ns.size
        self._A = np.zeros((self._dim_f, eq.Z_basis.num_modes))

        for i, (l, m, n) in enumerate(eq.Z_basis.modes):
            if m != 0:
                continue
            if (l // 2) % 2 == 0:
                j = np.argwhere(n == ns)
                self._A[j, i] = 1
            else:
                j = np.argwhere(n == ns)
                self._A[j, i] = -1

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute axis Z self-consistency errors.

        IE, the mismatch between the Fourier-Zernike basis evaluated at rho=0 and the
        Fourier series defining the equilibrium axis position

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            axis Z self-consistency errors.

        """
        f = jnp.dot(self._A, params["Z_lmn"]) - params["Za_n"]
        return f


class FixBoundaryR(FixParameters):
    """Boundary condition on the R boundary parameters.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Rb_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Rb_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of boundary modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the profile modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "R boundary error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="lcfs R",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.surface.R_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"Rb_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixBoundaryZ(FixParameters):
    """Boundary condition on the Z boundary parameters.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Zb_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Zb_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of boundary modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the surface modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "Z boundary error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="lcfs Z",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.surface.Z_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"Zb_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixLambdaGauge(FixParameters):
    """Fixes gauge freedom for lambda, which sets the flux surface avg of lambda to 0.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    name : str, optional
        Name of the objective function.

    """

    _units = "(rad)"
    _print_value_fmt = "lambda gauge error: "

    def __init__(
        self,
        eq,
        normalize=True,
        normalize_target=True,
        name="lambda gauge",
    ):
        if eq.sym:
            indices = False
        else:
            indices = np.where(
                np.logical_and(eq.L_basis.modes[:, 1] == 0, eq.L_basis.modes[:, 2] == 0)
            )[0]
        super().__init__(
            thing=eq,
            params={"L_lmn": indices},
            target=0,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixThetaSFL(FixParameters):
    """Fixes lambda=0 so that poloidal angle is the SFL poloidal angle.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    name : str, optional
        Name of the objective function.

    """

    _units = "(rad)"
    _print_value_fmt = "theta - theta SFL error: "

    def __init__(
        self,
        eq,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="theta SFL",
    ):
        super().__init__(
            thing=eq,
            params={"L_lmn": True},
            target=0,
            bounds=None,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixAxisR(FixParameters):
    """Fixes magnetic axis R coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Ra_n``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Ra_n``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of axis modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the axis modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "R axis error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="axis R",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.axis.R_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"Ra_n": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixAxisZ(FixParameters):
    """Fixes magnetic axis Z coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Za_n``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Za_n``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of axis modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the axis modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "Z axis error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="axis Z",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.axis.Z_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"Za_n": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixModeR(FixParameters):
    """Fixes Fourier-Zernike R coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.R_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.R_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the basis modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "Fixed-R modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="fix mode R",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.R_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"R_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixModeZ(FixParameters):
    """Fixes Fourier-Zernike Z coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Z_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Z_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the basis modes.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "Fixed-Z modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="fix mode Z",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.Z_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"Z_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["a"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixModeLambda(FixParameters):
    """Fixes Fourier-Zernike lambda coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : float, ndarray, optional
        Fourier-Zernike lambda coefficient target values.
        Must be broadcastable to Objective.dim_f.
        Defaults to ``target=eq.L_lmn``.
    bounds : tuple, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.L_lmn``.
    weight : float, ndarray, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to Objective.dim_f.
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix.
        len(target) = len(weight) = len(modes).
        If True/False uses all/none of the basis modes.
    name : str
        Name of the objective function.

    """

    _units = "(rad)"
    _print_value_fmt = "Fixed lambda modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        modes=True,
        name="fix mode lambda",
    ):
        if isinstance(modes, bool):
            indices = modes
        else:
            indices = np.array([], dtype=int)
            for mode in np.atleast_2d(modes):
                indices = np.append(indices, eq.L_basis.get_idx(*mode))
        super().__init__(
            thing=eq,
            params={"L_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixSumModesR(_FixedObjective):
    """Fixes a linear sum of Fourier-Zernike R coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.R_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.R_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    sum_weights : float, ndarray, optional
        Weights on the coefficients in the sum, should be same length as modes.
        Defaults to 1 i.e. target = 1*R_111 + 1*R_222...
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix sum of.
        len(weight) = len(modes).
        If True uses all of the Equilibrium's modes.
        Must be either True or specified as an array
    name : str, optional
        Name of the objective function.

    """

    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "Fixed-R sum modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        sum_weights=None,
        modes=True,
        name="Fix Sum Modes R",
    ):
        errorif(
            modes is None or modes is False,
            ValueError,
            f"modes kwarg must be specified or True with FixSumModesR got {modes}",
        )
        errorif(
            target is not None and np.asarray(target).size > 1,
            ValueError,
            "FixSumModesR only accepts 1 target value, please use multiple"
            + " FixSumModesR objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        errorif(
            bounds is not None and np.asarray(bounds)[0].size > 1,
            ValueError,
            "FixSumModesR only accepts 1 target value, please use multiple"
            + " FixSumModesR objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        self._modes = modes
        self._sum_weights = sum_weights
        self._target_from_user = setdefault(bounds, target)
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._modes is True:  # all modes
            modes = eq.R_basis.modes
            idx = np.arange(eq.R_basis.num_modes)
        else:  # specified modes
            modes = np.atleast_2d(self._modes)
            dtype = {
                "names": ["f{}".format(i) for i in range(3)],
                "formats": 3 * [modes.dtype],
            }
            _, idx, modes_idx = np.intersect1d(
                eq.R_basis.modes.astype(modes.dtype).view(dtype),
                modes.view(dtype),
                return_indices=True,
            )
            self._idx = idx
            # rearrange modes & weights to match order of eq.R_basis.modes and eq.R_lmn,
            # necessary so that the A matrix rows match up with the target b
            modes = np.atleast_2d(eq.R_basis.modes[idx, :])
            if self._sum_weights is not None:
                self._sum_weights = np.atleast_1d(self._sum_weights)
                self._sum_weights = self._sum_weights[modes_idx]
            if idx.size < modes.shape[0]:
                warnings.warn(
                    colored(
                        "Some of the given modes are not in the basis, "
                        + "these modes will not be fixed.",
                        "yellow",
                    )
                )
        if self._sum_weights is None:
            sum_weights = np.ones(modes.shape[0])
        else:
            sum_weights = np.atleast_1d(self._sum_weights)
        self._dim_f = 1

        self._A = np.zeros((1, eq.R_basis.num_modes))
        for i, (l, m, n) in enumerate(modes):
            j = eq.R_basis.get_idx(L=l, M=m, N=n)
            self._A[0, j] = sum_weights[i]

        self.target, self.bounds = self._parse_target_from_user(
            self._target_from_user,
            np.dot(sum_weights.T, eq.R_lmn[self._idx]),
            None,
            np.array([0]),
        )

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute Sum mode R errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed sum mode R errors.

        """
        f = jnp.dot(self._A, params["R_lmn"])
        return f


class FixSumModesZ(_FixedObjective):
    """Fixes a linear sum of Fourier-Zernike Z coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Z_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Z_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    sum_weights : float, ndarray, optional
        Weights on the coefficients in the sum, should be same length as modes.
        Defaults to 1 i.e. target = 1*Z_111 + 1*Z_222...
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix sum of.
        len(weight) = len(modes).
        If True uses all of the Equilibrium's modes.
        Must be either True or specified as an array
    name : str, optional
        Name of the objective function.

    """

    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(m)"
    _print_value_fmt = "Fixed-Z sum modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        sum_weights=None,
        modes=True,
        name="Fix Sum Modes Z",
    ):
        errorif(
            modes is None or modes is False,
            ValueError,
            f"modes kwarg must be specified or True with FixSumModesZ got {modes}",
        )
        errorif(
            target is not None and np.asarray(target).size > 1,
            ValueError,
            "FixSumModesZ only accepts 1 target value, please use multiple"
            + " FixSumModesZ objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        errorif(
            bounds is not None and np.asarray(bounds)[0].size > 1,
            ValueError,
            "FixSumModesZ only accepts 1 target value, please use multiple"
            + " FixSumModesZ objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        self._modes = modes
        self._sum_weights = sum_weights
        self._target_from_user = setdefault(bounds, target)
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._modes is True:  # all modes
            modes = eq.Z_basis.modes
            idx = np.arange(eq.Z_basis.num_modes)
        else:  # specified modes
            modes = np.atleast_2d(self._modes)
            dtype = {
                "names": ["f{}".format(i) for i in range(3)],
                "formats": 3 * [modes.dtype],
            }
            _, idx, modes_idx = np.intersect1d(
                eq.Z_basis.modes.astype(modes.dtype).view(dtype),
                modes.view(dtype),
                return_indices=True,
            )
            self._idx = idx
            # rearrange modes & weights to match order of eq.Z_basis.modes and eq.Z_lmn,
            # necessary so that the A matrix rows match up with the target b
            modes = np.atleast_2d(eq.Z_basis.modes[idx, :])
            if self._sum_weights is not None:
                self._sum_weights = np.atleast_1d(self._sum_weights)
                self._sum_weights = self._sum_weights[modes_idx]

            if idx.size < modes.shape[0]:
                warnings.warn(
                    colored(
                        "Some of the given modes are not in the basis, "
                        + "these modes will not be fixed.",
                        "yellow",
                    )
                )
        if self._sum_weights is None:
            sum_weights = np.ones(modes.shape[0])
        else:
            sum_weights = np.atleast_1d(self._sum_weights)
        self._dim_f = 1

        self._A = np.zeros((1, eq.Z_basis.num_modes))
        for i, (l, m, n) in enumerate(modes):
            j = eq.Z_basis.get_idx(L=l, M=m, N=n)
            self._A[0, j] = sum_weights[i]

        self.target, self.bounds = self._parse_target_from_user(
            self._target_from_user,
            np.dot(sum_weights.T, eq.Z_lmn[self._idx]),
            None,
            np.array([0]),
        )
        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute Sum mode Z errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed sum mode Z errors.

        """
        f = jnp.dot(self._A, params["Z_lmn"])
        return f


class FixSumModesLambda(_FixedObjective):
    """Fixes a linear sum of Fourier-Zernike lambda coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.L_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.L_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f.
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    sum_weight : float, ndarray, optional
        Weights on the coefficients in the sum, should be same length as modes.
        Defaults to 1 i.e. target = 1*L_111 + 1*L_222...
    modes : ndarray, optional
        Basis modes numbers [l,m,n] of Fourier-Zernike modes to fix sum of.
        len(weight) = len(modes).
        If True uses all of the Equilibrium's modes.
        Must be either True or specified as an array
    surface_label : float
        Surface to enforce boundary conditions on. Defaults to Equilibrium.surface.rho
    name : str
        Name of the objective function.

    """

    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(rad)"
    _print_value_fmt = "Fixed-lambda sum modes error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        sum_weights=None,
        modes=True,
        name="Fix Sum Modes lambda",
    ):
        errorif(
            modes is None or modes is False,
            ValueError,
            f"modes kwarg must be specified or True with FixSumModesLambda got {modes}",
        )
        errorif(
            target is not None and np.asarray(target).size > 1,
            ValueError,
            "FixSumModesLambda only accepts 1 target value, please use multiple"
            + " FixSumModesLambda objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        errorif(
            bounds is not None and np.asarray(bounds)[0].size > 1,
            ValueError,
            "FixSumModesLambda only accepts 1 target value, please use multiple"
            + " FixSumModesLambda objectives if you wish to have multiple"
            + " sets of constrained mode sums",
        )
        self._modes = modes
        self._sum_weights = sum_weights
        self._target_from_user = setdefault(bounds, target)
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    @execute_on_cpu
    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._modes is True:  # all modes
            modes = eq.L_basis.modes
            idx = np.arange(eq.L_basis.num_modes)
        else:  # specified modes
            modes = np.atleast_2d(self._modes)
            dtype = {
                "names": ["f{}".format(i) for i in range(3)],
                "formats": 3 * [modes.dtype],
            }
            _, idx, modes_idx = np.intersect1d(
                eq.L_basis.modes.astype(modes.dtype).view(dtype),
                modes.view(dtype),
                return_indices=True,
            )
            self._idx = idx
            # rearrange modes and weights to match order of eq.L_basis.modes
            # and eq.L_lmn,
            # necessary so that the A matrix rows match up with the target b
            modes = np.atleast_2d(eq.L_basis.modes[idx, :])
            if self._sum_weights is not None:
                self._sum_weights = np.atleast_1d(self._sum_weights)
                self._sum_weights = self._sum_weights[modes_idx]

            if idx.size < modes.shape[0]:
                warnings.warn(
                    colored(
                        "Some of the given modes are not in the basis, "
                        + "these modes will not be fixed.",
                        "yellow",
                    )
                )
        if self._sum_weights is None:
            sum_weights = np.ones(modes.shape[0])
        else:
            sum_weights = np.atleast_1d(self._sum_weights)
        self._dim_f = 1

        self._A = np.zeros((1, eq.L_basis.num_modes))
        for i, (l, m, n) in enumerate(modes):
            j = eq.L_basis.get_idx(L=l, M=m, N=n)
            self._A[0, j] = sum_weights[i]

        self.target, self.bounds = self._parse_target_from_user(
            self._target_from_user,
            np.dot(sum_weights.T, eq.L_lmn[self._idx]),
            None,
            np.array([0]),
        )

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute Sum mode lambda errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed sum mode lambda errors.

        """
        f = jnp.dot(self._A, params["L_lmn"])
        return f


class FixPressure(FixParameters):
    """Fixes pressure coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.p_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.p_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(Pa)"
    _print_value_fmt = "Fixed pressure profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed pressure",
    ):
        super().__init__(
            thing=eq,
            params={"p_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.pressure is None:
            raise RuntimeError(
                "Attempting to fix pressure on an Equilibrium with no "
                + "pressure profile assigned."
            )
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["p"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixAnisotropy(FixParameters):
    """Fixes anisotropic pressure coefficients.

    Parameters
    ----------
    eq : Equilibrium, optional
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.a_lmn``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.a_lmn``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str
        Name of the objective function.

    """

    _units = "(dimensionless)"
    _print_value_fmt = "Fixed anisotropy profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed anisotropy",
    ):
        super().__init__(
            thing=eq,
            params={"a_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.anisotropy is None:
            raise RuntimeError(
                "Attempting to fix anisotropy on an Equilibrium with no "
                + "anisotropy profile assigned."
            )
        super().build(use_jit=use_jit, verbose=verbose)


class FixIota(FixParameters):
    """Fixes rotational transform coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.i_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.i_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices.
        corresponding to knots for a SplineProfile).
        Must len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(dimensionless)"
    _print_value_fmt = "Fixed iota profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed iota",
    ):
        super().__init__(
            thing=eq,
            params={"i_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.iota is None:
            raise RuntimeError(
                "Attempting to fix iota on an Equilibrium with no "
                + "iota profile assigned."
            )
        super().build(use_jit=use_jit, verbose=verbose)


class FixCurrent(FixParameters):
    """Fixes toroidal current profile coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.c_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.c_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(A)"
    _print_value_fmt = "Fixed current profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed current",
    ):
        super().__init__(
            thing=eq,
            params={"c_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.current is None:
            raise RuntimeError(
                "Attempting to fix current on an Equilibrium with no "
                + "current profile assigned."
            )
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["I"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixElectronTemperature(FixParameters):
    """Fixes electron temperature profile coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Te_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Te_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(eV)"
    _print_value_fmt = "Fixed electron temperature profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed electron temperature",
    ):
        super().__init__(
            thing=eq,
            params={"Te_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.electron_temperature is None:
            raise RuntimeError(
                "Attempting to fix electron temperature on an Equilibrium with no "
                + "electron temperature profile assigned."
            )
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["T"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixElectronDensity(FixParameters):
    """Fixes electron density profile coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.ne_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.ne_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    profile : Profile, optional
        Profile containing the radial modes to evaluate at.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m^-3)"
    _print_value_fmt = "Fixed electron density profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed electron density",
    ):
        super().__init__(
            thing=eq,
            params={"ne_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.electron_density is None:
            raise RuntimeError(
                "Attempting to fix electron density on an Equilibrium with no "
                + "electron density profile assigned."
            )
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["n"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixIonTemperature(FixParameters):
    """Fixes ion temperature profile coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Ti_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Ti_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(eV)"
    _print_value_fmt = "Fixed ion temperature profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed ion temperature",
    ):
        super().__init__(
            thing=eq,
            params={"Ti_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.ion_temperature is None:
            raise RuntimeError(
                "Attempting to fix ion temperature on an Equilibrium with no "
                + "ion temperature profile assigned."
            )
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["T"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixAtomicNumber(FixParameters):
    """Fixes effective atomic number profile coefficients.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Defaults to ``target=eq.Zeff_l``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Defaults to ``target=eq.Zeff_l``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    indices : ndarray or bool, optional
        indices of the Profile.params array to fix.
        (e.g. indices corresponding to modes for a PowerSeriesProfile or indices
        corresponding to knots for a SplineProfile).
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the Profile.params indices.
    name : str, optional
        Name of the objective function.

    """

    _units = "(dimensionless)"
    _print_value_fmt = "Fixed atomic number profile error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed atomic number",
    ):
        super().__init__(
            thing=eq,
            params={"Zeff_l": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if eq.atomic_number is None:
            raise RuntimeError(
                "Attempting to fix atomic number on an Equilibrium with no "
                + "atomic_number profile assigned."
            )
        super().build(use_jit=use_jit, verbose=verbose)


class FixPsi(FixParameters):
    """Fixes total toroidal magnetic flux within the last closed flux surface.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f. Default is ``target=eq.Psi``.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Default is ``target=eq.Psi``.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    name : str, optional
        Name of the objective function.

    """

    _units = "(Wb)"
    _print_value_fmt = "Fixed Psi error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed Psi",
    ):
        super().__init__(
            thing=eq,
            params={"Psi": True},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["Psi"]
        super().build(use_jit=use_jit, verbose=verbose)


class FixCurveShift(FixParameters):
    """Fixes Curve.shift attribute, which is redundant with other Curve params.

    Parameters
    ----------
    curve : Curve
        Curve that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    name : str, optional
        Name of the objective function.

    """

    _units = "(m)"
    _print_value_fmt = "Fixed shift error: "

    def __init__(
        self,
        curve,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed shift",
    ):
        super().__init__(
            thing=curve,
            params={"shift": True},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixCurveRotation(FixParameters):
    """Fixes Curve.rotmat attribute, which is redundant with other Curve params.

    Parameters
    ----------
    curve : Curve
        Curve that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    name : str, optional
        Name of the objective function.

    """

    _units = "(rad)"
    _print_value_fmt = "Fixed rotation error: "

    def __init__(
        self,
        curve,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed rotation",
    ):
        super().__init__(
            thing=curve,
            params={"rotmat": True},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixCoilCurrent(FixParameters):
    """Fixes current(s) in a Coil or CoilSet.

    Parameters
    ----------
    coil : Coil
        Coil(s) that will be optimized to satisfy the Objective.
    target : dict of {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Should have the same tree structure as coil.params.
        Default is ``target=coil.current``.
    bounds : tuple of dict {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Should have the same tree structure as coil.params.
        Default is ``target=coil.current``.
    weight : dict of {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Should be a scalar or have the same tree structure as coil.params.
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : nested list of bool, optional
        Pytree of bool specifying which coil currents to fix.
        See the example for how to use this on a mixed coil set.
        If True/False fixes all/none of the coil currents.
    name : str, optional
        Name of the objective function.

    Examples
    --------
    .. code-block:: python

        import numpy as np
        from desc.coils import (
            CoilSet, FourierPlanarCoil, FourierRZCoil, FourierXYZCoil, MixedCoilSet
        )
        from desc.objectives import FixCoilCurrent

        # toroidal field coil set with 4 coils
        tf_coil = FourierPlanarCoil(
            current=3, center=[2, 0, 0], normal=[0, 1, 0], r_n=[1]
        )
        tf_coilset = CoilSet.linspaced_angular(tf_coil, n=4)
        # vertical field coil set with 3 coils
        vf_coil = FourierRZCoil(current=-1, R_n=3, Z_n=-1)
        vf_coilset = CoilSet.linspaced_linear(
            vf_coil, displacement=[0, 0, 2], n=3, endpoint=True
        )
        # another single coil
        xyz_coil = FourierXYZCoil(current=2)
        # full coil set with TF coils, VF coils, and other single coil
        full_coilset = MixedCoilSet((tf_coilset, vf_coilset, xyz_coil))

        # fix the current of the 1st & 3rd TF coil
        # fix none of the currents in the VF coil set
        # fix the current of the other coil
        obj = FixCoilCurrent(
            full_coilset, indices=[[True, False, True, False], False, True]
        )

    """

    _units = "(A)"
    _print_value_fmt = "Fixed coil current error: "

    def __init__(
        self,
        coil,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed coil current",
    ):
        indices = tree_map(lambda idx: {"current": idx}, indices)
        super().__init__(
            thing=coil,
            params=indices,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        coil = self.things[0]
        if self._normalize:
            params = tree_leaves(
                coil.params_dict, is_leaf=lambda x: isinstance(x, dict)
            )
            mean_current = np.mean([np.abs(param["current"]) for param in params])
            self._normalization = np.max((mean_current, 1))
        super().build(use_jit=use_jit, verbose=verbose)


class FixSumCoilCurrent(FixCoilCurrent):
    """Fixes the sum of coil current(s) in a Coil or CoilSet.

    NOTE: When using this objective, take care in knowing the signs of the current in
    the coils and the orientations of the coils. It is possible for coils with the same
    signs of their current to have currents flowing in differing directions in physical
    space due to the orientation of the coils.

    Parameters
    ----------
    coil : Coil
        Coil(s) that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
        Default is the objective value for the coil.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Default is to use the target instead.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    indices : nested list of bool, optional
        Pytree of bool specifying which coil currents to sum together.
        See the example for how to use this on a mixed coil set.
        If True/False sums all/none of the coil currents.
    name : str, optional
        Name of the objective function.

    Examples
    --------
    .. code-block:: python

        import numpy as np
        from desc.coils import (
            CoilSet, FourierPlanarCoil, FourierRZCoil, FourierXYZCoil, MixedCoilSet
        )
        from desc.objectives import FixSumCoilCurrent

        # toroidal field coil set with 4 coils
        tf_coil = FourierPlanarCoil(
            current=3, center=[2, 0, 0], normal=[0, 1, 0], r_n=[1]
        )
        tf_coilset = CoilSet.linspaced_angular(tf_coil, n=4)
        # vertical field coil set with 3 coils
        vf_coil = FourierRZCoil(current=-1, R_n=3, Z_n=-1)
        vf_coilset = CoilSet.linspaced_linear(
            vf_coil, displacement=[0, 0, 2], n=3, endpoint=True
        )
        # another single coil
        xyz_coil = FourierXYZCoil(current=2)
        # full coil set with TF coils, VF coils, and other single coil
        full_coilset = MixedCoilSet((tf_coilset, vf_coilset, xyz_coil))

        # equilibrium G(rho=1) determines the necessary net poloidal current through
        # the coils (as dictated by Ampere's law)
        # the sign convention is positive poloidal current flows up through the torus
        # hole
        grid_at_surf = LinearGrid(rho=1.0, M=eq.M_grid, N=eq.N_grid)
        G_tot = 2*jnp.pi*eq.compute("G", grid=grid_at_surf)["G"][0] / mu_0

        # to use this objective to satisfy Ampere's law for the targeted equilibrium,
        # only coils that link the equilibrium poloidally should be included in the sum,
        # which is the TF coil set and the FourierXYZ coil, but not the VF coil set
        obj = FixSumCoilCurrent(full_coilset, indices=[True, False, True], target=G_tot)

    """

    _scalar = True
    _linear = True
    _fixed = False
    _units = "(A)"
    _print_value_fmt = "Summed coil current error: "

    def __init__(
        self,
        coil,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="summed coil current",
    ):
        self._default_target = False
        if target is None and bounds is None:
            self._default_target = True
        super().__init__(
            coil=coil,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            indices=indices,
            name=name,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        super().build(use_jit=use_jit, verbose=verbose)
        self._dim_f = 1
        if self._default_target:
            self.update_target(thing=self.things[0])

    def compute(self, params, constants=None):
        """Compute sum of coil currents.

        Parameters
        ----------
        params : list of dict
            List of dictionaries of degrees of freedom, eg CoilSet.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Sum of coil currents.

        """
        return jnp.sum(
            jnp.concatenate(
                [
                    jnp.atleast_1d(param[idx])
                    for param, idx in zip(tree_leaves(params), self._indices)
                ]
            )
        )


class FixOmniWell(FixParameters):
    """Fixes OmnigenousField.B_lm coefficients.

    Parameters
    ----------
    field : OmnigenousField
        Field that will be optimized to satisfy the Objective.
    target : float, optional
        Target value(s) of the objective. If None, uses field value.
    bounds : tuple, optional
        Lower and upper bounds on the objective. Overrides target.
    weight : float, optional
        Weighting to apply to the Objective, relative to other Objectives.
    normalize : bool
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool
        Whether target should be normalized before comparing to computed values.
        if `normalize` is `True` and the target is in physical units, this should also
        be set to True.
    indices : ndarray or bool, optional
        indices of the field.B_lm array to fix.
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the field.B_lm indices.
    name : str
        Name of the objective function.

    """

    _units = "(T)"
    _print_value_fmt = "Fixed omnigenity well error: "

    def __init__(
        self,
        field,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed omnigenity well",
    ):
        super().__init__(
            thing=field,
            params={"B_lm": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixOmniMap(FixParameters):
    """Fixes OmnigenousField.x_lmn coefficients.

    Parameters
    ----------
    field : OmnigenousField
        Field that will be optimized to satisfy the Objective.
    target : float, optional
        Target value(s) of the objective. If None, uses field value.
    bounds : tuple, optional
        Lower and upper bounds on the objective. Overrides target.
    weight : float, optional
        Weighting to apply to the Objective, relative to other Objectives.
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    indices : ndarray or bool, optional
        indices of the field.x_lmn array to fix.
        Must have len(target) = len(weight) = len(indices).
        If True/False uses all/none of the field.x_lmn indices.
    name : str
        Name of the objective function.

    """

    _units = "(rad)"
    _print_value_fmt = "Fixed omnigenity map error: "

    def __init__(
        self,
        field,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        indices=True,
        name="fixed omnigenity map",
    ):
        super().__init__(
            thing=field,
            params={"x_lmn": indices},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixOmniBmax(_FixedObjective):
    """Ensures the B_max contour is straight in Boozer coordinates.

    Parameters
    ----------
    field : OmnigenousField
        Field that will be optimized to satisfy the Objective.
    target : float, optional
        Target value(s) of the objective. If None, uses field value.
    bounds : tuple, optional
        Lower and upper bounds on the objective. Overrides target.
    weight : float, optional
        Weighting to apply to the Objective, relative to other Objectives.
    normalize : bool, optional
        Has no effect for this objective.
    normalize_target : bool, optional
        Has no effect for this objective.
    name : str
        Name of the objective function.

    """

    _fixed = False  # not "diagonal", since it is fixing a sum
    _units = "(rad)"
    _print_value_fmt = "Fixed omnigenity B_max error: "

    def __init__(
        self,
        field,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed omnigenity B_max",
    ):
        self._target_from_user = setdefault(bounds, target)
        super().__init__(
            things=field,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    @execute_on_cpu
    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        field = self.things[0]

        basis = field.x_basis
        self._dim_f = int(basis.num_modes / (basis.M + 1))

        self._A = np.zeros((self._dim_f, basis.num_modes))
        m0_modes = basis.modes[np.nonzero(basis.modes[:, 1] == 0)[0]]
        for i, (l, m, n) in enumerate(m0_modes):
            idx_0 = np.nonzero((basis.modes == [l, m, n]).all(axis=1))[0]
            idx_m = np.nonzero(
                np.logical_and(
                    (basis.modes[:, (0, 2)] == [l, n]).all(axis=1),
                    np.logical_and((basis.modes[:, 1] % 2 == 0), basis.modes[:, 1] > 0),
                )
            )[0]
            mm = basis.modes[idx_m, 1]
            self._A[i, idx_0] = 1
            self._A[i, idx_m] = (mm % 2 - 1) * (mm % 4 - 1)

        self.target, self.bounds = self._parse_target_from_user(
            self._target_from_user, 0, None, np.array([0])
        )

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed omnigenity B_max error.

        Parameters
        ----------
        params : dict
            Dictionary of field degrees of freedom, eg OmnigenousField.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed omnigenity B_max error.

        """
        f = jnp.dot(self._A, params["x_lmn"])
        return f


class FixSheetCurrent(FixParameters):
    """Fixes the sheet current parameters of a free-boundary equilibrium.

    Note: this constraint is automatically applied when needed, and does not need to be
    included by the user.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
        Defaults to the equilibrium sheet current parameters.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f.
        Default is to use target.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
    name : str, optional
        Name of the objective function.

    """

    _units = "(~)"
    _print_value_fmt = "Fixed sheet current error: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="fixed sheet current",
    ):
        super().__init__(
            thing=eq,
            params={"I": True, "G": True, "Phi_mn": True},
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )


class FixNearAxisR(_FixedObjective):
    """Fixes an equilibrium's near-axis behavior in R to specified order.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    order : {0,1,2}
        order (in rho) of near-axis behavior to constrain
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    target : Qsc, optional
        pyQSC Qsc object describing the NAE solution to fix the equilibrium's
        near-axis behavior to. If None, will fix the equilibrium's current near
        axis behavior.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
        Unused by this objective
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
        Unused by this objective
    name : str, optional
        Name of the objective function.

    """

    _static_attrs = ["_nae_eq"]
    _target_arg = "R_lmn"
    _fixed = False  # not "diagonal", since its fixing a sum
    _units = "(m)"
    _print_value_fmt = "Fixed-Near-Axis R Behavior error: "

    def __init__(
        self,
        eq,
        order=1,
        N=None,
        target=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="Fix Near Axis R Behavior",
    ):
        self._nae_eq = target
        self._eq = eq
        self._order = order
        self._N = N
        super().__init__(
            things=eq,
            target=None,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        from .getters import _get_NAE_constraints

        eq = self.things[0]
        cons = _get_NAE_constraints(
            eq, self._nae_eq, order=self._order, N=self._N, fix_lambda=False
        )
        for con in cons:
            if isinstance(con, FixSumModesR) or isinstance(con, AxisRSelfConsistency):
                con.build(use_jit=use_jit, verbose=0)
        cons_that_fix_near_axis = [con for con in cons if isinstance(con, FixSumModesR)]
        self._A = (
            np.stack([con._A for con in cons_that_fix_near_axis]).squeeze()
            if cons_that_fix_near_axis
            else None
        )
        # add the axis constraint last if it is in cons
        axis_con = None
        for con in cons:
            if isinstance(con, AxisRSelfConsistency):
                self._A = (
                    np.vstack([self._A, con._A]).squeeze()
                    if np.any(self._A)
                    else con._A
                )
                axis_con = con

        self._target = (
            np.concatenate([con.target for con in cons_that_fix_near_axis])
            if cons_that_fix_near_axis
            else None
        )
        if axis_con:
            if self._nae_eq is not None:
                # use NAE axis as target
                axis = FourierRZCurve(
                    R_n=np.concatenate(
                        (np.flipud(self._nae_eq.rs[1:]), self._nae_eq.rc)
                    ),
                    Z_n=np.concatenate(
                        (np.flipud(self._nae_eq.zs[1:]), self._nae_eq.zc)
                    ),
                    NFP=self._nae_eq.nfp,
                )
                axis.change_resolution(N=self._eq.N)
                axis_target = axis.R_n
            else:  # else use eq axis a target
                axis_target = self._eq.Ra_n
            self._target = (
                np.append(self._target, axis_target).squeeze()
                if np.any(self._target)
                else axis_target
            )
        self._dim_f = self.target.size
        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed near axis R behavior errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed near axis behavior errors.

        """
        f = jnp.dot(self._A, params["R_lmn"]).squeeze()

        return f


class FixNearAxisZ(_FixedObjective):
    """Fixes an equilibrium's near-axis behavior in Z to specified order.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    order : {0,1,2}
        order (in rho) of near-axis behavior to constrain
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    target : Qsc, optional
        pyQSC Qsc object describing the NAE solution to fix the equilibrium's
        near-axis behavior to. If None, will fix the equilibrium's current near
        axis behavior.
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
        Unused by this objective
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
        Unused by this objective
    name : str, optional
        Name of the objective function.

    """

    _static_attrs = ["_nae_eq"]
    _target_arg = "Z_lmn"
    _fixed = False  # not "diagonal", since its fixing a sum
    _units = "(m)"
    _print_value_fmt = "Fixed-Near-Axis Z Behavior error: "

    def __init__(
        self,
        eq,
        order=1,
        N=None,
        target=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="Fix Near Axis Z Behavior",
    ):
        self._nae_eq = target
        self._eq = eq
        self._order = order
        self._N = N
        super().__init__(
            things=eq,
            target=None,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        from .getters import _get_NAE_constraints

        eq = self.things[0]
        cons = _get_NAE_constraints(
            eq, self._nae_eq, order=self._order, N=self._N, fix_lambda=False
        )
        for con in cons:
            if isinstance(con, FixSumModesZ) or isinstance(con, AxisZSelfConsistency):
                con.build(use_jit=use_jit, verbose=0)
        cons_that_fix_near_axis = [con for con in cons if isinstance(con, FixSumModesZ)]
        self._A = (
            np.stack([con._A for con in cons_that_fix_near_axis]).squeeze()
            if cons_that_fix_near_axis
            else None
        )
        # add the axis constraint last if it is in cons
        axis_con = None
        for con in cons:
            if isinstance(con, AxisZSelfConsistency):
                self._A = (
                    np.vstack([self._A, con._A]).squeeze()
                    if np.any(self._A)
                    else con._A
                )
                axis_con = con
        self._target = (
            np.concatenate([con.target for con in cons_that_fix_near_axis])
            if cons_that_fix_near_axis
            else None
        )
        if axis_con:
            if self._nae_eq is not None:
                # use NAE axis as target
                axis = FourierRZCurve(
                    R_n=np.concatenate(
                        (np.flipud(self._nae_eq.rs[1:]), self._nae_eq.rc)
                    ),
                    Z_n=np.concatenate(
                        (np.flipud(self._nae_eq.zs[1:]), self._nae_eq.zc)
                    ),
                    NFP=self._nae_eq.nfp,
                )
                axis.change_resolution(N=self._eq.N)
                axis_target = axis.Z_n
            else:  # else use eq axis a target
                axis_target = self._eq.Za_n
            self._target = (
                np.append(self._target, axis_target).squeeze()
                if np.any(self._target)
                else axis_target
            )

        self._dim_f = self.target.size

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed near axis Z behavior errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed near axis behavior errors.

        """
        f = jnp.dot(self._A, params["Z_lmn"]).squeeze()

        return f


class FixNearAxisLambda(_FixedObjective):
    """Fixes an equilibrium's near-axis behavior in lambda to specified order.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    order : int
        order (in rho) of near-axis lambda behavior to constrain
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    target : Qsc, optional
        pyQSC Qsc object describing the NAE solution to fix the equilibrium's
        near-axis behavior to. If None, will fix the equilibrium's current near
        axis behavior.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f
        Unused for this objective, as target will be automatically
        set according to the ``nae_eq``
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
        Unused by this objective
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True.
        Unused by this objective
    name : str, optional
        Name of the objective function.

    """

    _static_attrs = ["_nae_eq"]
    _target_arg = "L_lmn"
    _fixed = False  # not "diagonal", since its fixing a sum
    _units = "(dimensionless)"
    _print_value_fmt = "Fixed-Near-Axis Lambda Behavior error: "

    def __init__(
        self,
        eq,
        order=1,
        N=None,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        name="Fix Near Axis Lambda Behavior",
    ):
        self._nae_eq = target
        self._order = order
        self._N = N
        self._target_from_user = setdefault(bounds, None)
        super().__init__(
            things=eq,
            target=None,
            bounds=bounds,
            weight=weight,
            name=name,
            normalize=normalize,
            normalize_target=normalize_target,
        )

    def build(self, use_jit=False, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        from .getters import _get_NAE_constraints

        eq = self.things[0]
        cons = _get_NAE_constraints(
            eq,
            self._nae_eq,
            order=self._order,
            N=self._N,
            fix_lambda=self._order,
        )
        for con in cons:
            if isinstance(con, FixSumModesLambda):
                con.build(use_jit=use_jit, verbose=0)
        self._A = np.vstack(
            [con._A for con in cons if isinstance(con, FixSumModesLambda)]
        )
        self._target = np.concatenate(
            [con.target for con in cons if isinstance(con, FixSumModesLambda)]
        )

        self._dim_f = self.target.size

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fixed near axis Lambda behavior errors.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium degrees of freedom, eg Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        f : ndarray
            Fixed near axis Lambda behavior errors.

        """
        f = jnp.dot(self._A, params["L_lmn"]).squeeze()
        return f



"""Tests for linear constraints and objectives."""

import numpy as np
import pytest
from qsc import Qsc

import desc.examples
from desc.backend import jnp
from desc.equilibrium import Equilibrium
from desc.geometry import FourierRZToroidalSurface
from desc.io import load
from desc.magnetic_fields import OmnigenousField
from desc.objectives import (
    AspectRatio,
    AxisRSelfConsistency,
    AxisZSelfConsistency,
    BoundaryRSelfConsistency,
    BoundaryZSelfConsistency,
    FixAtomicNumber,
    FixAxisR,
    FixAxisZ,
    FixBoundaryR,
    FixBoundaryZ,
    FixCoilCurrent,
    FixCurrent,
    FixElectronDensity,
    FixElectronTemperature,
    FixIonTemperature,
    FixIota,
    FixLambdaGauge,
    FixModeLambda,
    FixModeR,
    FixModeZ,
    FixNearAxisLambda,
    FixNearAxisR,
    FixNearAxisZ,
    FixOmniBmax,
    FixOmniMap,
    FixOmniWell,
    FixParameters,
    FixPressure,
    FixPsi,
    FixSumCoilCurrent,
    FixSumModesLambda,
    FixSumModesR,
    FixSumModesZ,
    FixThetaSFL,
    GenericObjective,
    LinearObjectiveFromUser,
    ObjectiveFunction,
    get_equilibrium_objective,
    get_fixed_axis_constraints,
    get_fixed_boundary_constraints,
    get_NAE_constraints,
)
from desc.objectives.utils import factorize_linear_constraints
from desc.profiles import PowerSeriesProfile

# TODO (#1348): check for all bdryR things if work when False is passed in
# bc empty array indexed will lead to an error


@pytest.mark.unit
def test_LambdaGauge_sym(DummyStellarator):
    """Test that lambda is fixed correctly for symmetric equilibrium."""
    # symmetric cases automatically satisfy gauge freedom, no constraint needed.
    eq = load(load_from=str(DummyStellarator["output_path"]), file_format="hdf5")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(L=2, M=1, N=1)
    lam_con = FixLambdaGauge(eq)
    lam_con.build()
    # should have no indices to fix
    assert lam_con._params["L_lmn"].size == 0


@pytest.mark.unit
def test_LambdaGauge_asym():
    """Test that lambda gauge is fixed correctly for asymmetric equilibrium."""
    # just testing the gauge condition
    inputs = {
        "sym": False,
        "NFP": 3,
        "Psi": 1.0,
        "L": 2,
        "M": 2,
        "N": 1,
        "pressure": np.array([[0, 1e4], [2, -2e4], [4, 1e4]]),
        "iota": np.array([[0, 0.5], [2, 0.5]]),
        "surface": np.array(
            [
                [0, 0, 0, 3, 0],
                [0, 1, 0, 1, 0],
                [0, -1, 0, 0, -1],
                [0, 1, 1, 0.3, 0],
                [0, -1, -1, 0.3, 0],
                [0, 1, -1, 0, -0.3],
                [0, -1, 1, 0, 0.3],
            ],
        ),
        "axis": np.array([[-1, 0, -0.2], [0, 3.4, 0], [1, 0.2, 0]]),
    }
    eq = Equilibrium(**inputs)
    lam_con = FixLambdaGauge(eq)
    lam_con.build()

    indices = np.where(
        np.logical_and(eq.L_basis.modes[:, 1] == 0, eq.L_basis.modes[:, 2] == 0)
    )[0]
    np.testing.assert_allclose(indices, lam_con._params["L_lmn"])


@pytest.mark.regression
@pytest.mark.solve
def test_bc_on_interior_surfaces():
    """Test applying boundary conditions on internal surface."""
    surf = FourierRZToroidalSurface(rho=0.5)
    iota = PowerSeriesProfile([1, 0, 0.5])
    eq = Equilibrium(L=4, M=4, N=0, surface=surf, iota=iota)
    eq.solve(verbose=0)
    surf5 = eq.get_surface_at(0.5)

    np.testing.assert_allclose(surf.R_lmn, surf5.R_lmn, atol=1e-12)
    np.testing.assert_allclose(surf.Z_lmn, surf5.Z_lmn, atol=1e-12)


@pytest.mark.unit
def test_constrain_bdry_with_only_one_mode():
    """Test Fixing boundary with a surface with only one mode in its basis."""
    eq = Equilibrium()
    FixZ = BoundaryZSelfConsistency(eq=eq)
    try:
        FixZ.build()
    except Exception:
        pytest.fail(
            "Error encountered when attempting to constrain surface with"
            + " only one mode in its basis"
        )


@pytest.mark.unit
def test_constrain_asserts():
    """Test error checking for incompatible constraints."""
    eqi = Equilibrium(iota=PowerSeriesProfile(0, 0), pressure=PowerSeriesProfile(0, 0))
    eqc = Equilibrium(current=PowerSeriesProfile(1))
    obj_i = get_equilibrium_objective(eqi, "force")
    obj_c = get_equilibrium_objective(eqc, "force")
    obj_i.build()
    obj_c.build()
    # nonexistent toroidal current can't be constrained
    with pytest.raises(RuntimeError):
        con = FixCurrent(eq=eqi)
        con.build()
    # nonexistent rotational transform can't be constrained
    with pytest.raises(RuntimeError):
        con = FixIota(eq=eqc)
        con.build()
    # toroidal current and rotational transform can't be constrained simultaneously
    with pytest.raises(ValueError):
        con = (FixCurrent(eq=eqi), FixIota(eq=eqi))
        eqi.solve(constraints=con)
    # cannot use two incompatible constraints
    with pytest.raises(AssertionError):
        con1 = FixCurrent(target=eqc.c_l, eq=eqc)
        con2 = FixCurrent(target=eqc.c_l + 1, eq=eqc)
        con = ObjectiveFunction((con1, con2))
        con.build()
        _ = factorize_linear_constraints(obj_c, con)
    # if only slightly off, should raise only a warning
    with pytest.warns(UserWarning):
        con1 = FixCurrent(target=eqc.c_l, eq=eqc)
        con2 = FixCurrent(target=eqc.c_l * (1 + 1e-9), eq=eqc)
        con = ObjectiveFunction((con1, con2))
        con.build()
        _ = factorize_linear_constraints(obj_c, con)


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_mode_solve():
    """Test solving an equilibrium with a fixed mode constraint."""
    # Reset DSHAPE to initial guess, fix a mode, and then resolve
    # and check that the mode stayed fix
    L = 1
    M = 1
    eq = desc.examples.get("DSHAPE")
    eq.set_initial_guess()
    fixR = FixModeR(
        eq=eq, modes=np.array([L, M, 0])
    )  # no target supplied, so defaults to the eq's current R_LMN coeff
    fixZ = FixModeZ(
        eq=eq, modes=np.array([L, -M, 0])
    )  # no target supplied, so defaults to the eq's current Z_LMN coeff
    orig_R_val = eq.R_lmn[eq.R_basis.get_idx(L=L, M=M, N=0)]
    orig_Z_val = eq.Z_lmn[eq.Z_basis.get_idx(L=L, M=-M, N=0)]

    constraints = (
        FixLambdaGauge(eq=eq),
        FixPressure(eq=eq),
        FixIota(eq=eq),
        FixPsi(eq=eq),
        FixBoundaryR(eq=eq),
        FixBoundaryR(
            eq=eq, modes=[0, 0, 0]
        ),  # add a degenerate constraint to test fix of GH #1297 for lsq-exact
        FixBoundaryZ(eq=eq),
        fixR,
        fixZ,
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )
    np.testing.assert_almost_equal(
        orig_R_val, eq.R_lmn[eq.R_basis.get_idx(L=L, M=M, N=0)]
    )
    np.testing.assert_almost_equal(
        orig_Z_val, eq.Z_lmn[eq.Z_basis.get_idx(L=L, M=-M, N=0)]
    )


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_modes_solve():
    """Test solving an equilibrium with fixed sum modes constraint."""
    # Reset DSHAPE to initial guess, fix sum modes, and then resolve
    # and check that the mode sum stayed fix
    modes_R = np.array([[1, 1, 0], [2, 2, 0]])
    modes_Z = np.array([[1, -1, 0], [2, -2, 0]])

    eq = desc.examples.get("DSHAPE")
    eq.set_initial_guess()
    fixR = FixSumModesR(
        eq=eq, modes=modes_R, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum
    fixZ = FixSumModesZ(
        eq=eq, modes=modes_Z, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum
    fixLambda = FixSumModesLambda(
        eq=eq, modes=modes_Z, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum

    orig_R_val = (
        eq.R_lmn[eq.R_basis.get_idx(L=modes_R[0, 0], M=modes_R[0, 1], N=0)]
        + 2 * eq.R_lmn[eq.R_basis.get_idx(L=modes_R[1, 0], M=modes_R[1, 1], N=0)]
    )
    orig_Z_val = (
        eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    orig_Lambda_val = (
        eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    constraints = (
        FixLambdaGauge(eq=eq),
        FixPressure(eq=eq),
        FixIota(eq=eq),
        FixPsi(eq=eq),
        FixBoundaryR(eq=eq),
        FixBoundaryZ(eq=eq),
        fixR,
        fixZ,
        fixLambda,
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )

    new_R_val = (
        eq.R_lmn[eq.R_basis.get_idx(L=modes_R[0, 0], M=modes_R[0, 1], N=0)]
        + 2 * eq.R_lmn[eq.R_basis.get_idx(L=modes_R[1, 0], M=modes_R[1, 1], N=0)]
    )
    new_Z_val = (
        eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    new_Lambda_val = (
        eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )

    np.testing.assert_almost_equal(orig_R_val, new_R_val)
    np.testing.assert_almost_equal(orig_Z_val, new_Z_val)
    np.testing.assert_almost_equal(orig_Lambda_val, new_Lambda_val)


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_axis_and_theta_SFL_solve():
    """Test solving an equilibrium with a fixed axis and theta SFL constraint."""
    # also tests zero lambda solve
    eq = Equilibrium()

    orig_R_val = eq.axis.R_n
    orig_Z_val = eq.axis.Z_n

    constraints = (FixThetaSFL(eq=eq),) + get_NAE_constraints(
        eq, None, order=0, fix_lambda=False
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )

    np.testing.assert_allclose(orig_R_val, eq.axis.R_n, atol=1e-14)
    np.testing.assert_allclose(orig_Z_val, eq.axis.Z_n, atol=1e-14)
    np.testing.assert_allclose(np.zeros_like(eq.L_lmn), eq.L_lmn, atol=1e-14)


@pytest.mark.unit
def test_factorize_linear_constraints_asserts():
    """Test error checking for factorize_linear_constraints."""
    eq = Equilibrium()
    surf = eq.get_surface_at(rho=1)
    objective = get_equilibrium_objective(eq, "force")
    objective.build()

    # nonlinear constraint
    constraint = ObjectiveFunction(AspectRatio(eq=eq))
    constraint.build(verbose=0)
    with pytest.raises(ValueError):
        _ = factorize_linear_constraints(objective, constraint)

    # bounds instead of target
    constraint = ObjectiveFunction(get_fixed_boundary_constraints(eq=eq))
    constraint.build(verbose=0)
    constraint.objectives[3].bounds = (0, 1)
    with pytest.raises(ValueError):
        _ = factorize_linear_constraints(objective, constraint)

    # constraining a foreign thing
    constraint = ObjectiveFunction(FixParameters(surf))
    constraint.build(verbose=0)
    with pytest.raises(UserWarning):
        _ = factorize_linear_constraints(objective, constraint)


@pytest.mark.unit
def test_build_init():
    """Ensure that passing an equilibrium to init builds the objective correctly.

    Test for GH issue #378.
    """
    eq = Equilibrium(M=3, N=1)

    # initialize the constraints without building
    fbR1 = FixBoundaryR(eq=eq)
    fbZ1 = FixBoundaryZ(eq=eq)
    for obj in (fbR1, fbZ1):
        obj.build()

    xz = {key: np.zeros_like(val) for key, val in eq.params_dict.items()}
    arg = "Rb_lmn"
    A = fbR1.jac_scaled(xz)[0][arg]
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.R_basis.num_modes, eq.surface.R_basis.num_modes)

    arg = "Zb_lmn"
    A = fbZ1.jac_scaled(xz)[0][arg]
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.Z_basis.num_modes, eq.surface.Z_basis.num_modes)


@pytest.mark.unit
def test_kinetic_constraints():
    """Make sure errors are raised when trying to constrain nonexistent profiles."""
    eqp = Equilibrium(L=3, M=3, N=3, pressure=np.array([1, 0, -1]))
    eqk = Equilibrium(
        L=3,
        M=3,
        N=3,
        electron_temperature=np.array([1, 0, -1]),
        electron_density=np.array([2, 0, -2]),
    )
    pcon = (FixPressure(eq=eqk),)
    kcon = (
        FixAtomicNumber(eq=eqp),
        FixElectronDensity(eq=eqp),
        FixElectronTemperature(eq=eqp),
        FixIonTemperature(eq=eqp),
    )
    for con in pcon:
        with pytest.raises(RuntimeError):
            con.build()
    for con in kcon:
        with pytest.raises(RuntimeError):
            con.build()


@pytest.mark.unit
def test_correct_indexing_passed_modes():
    """Test indexing when passing in specified modes, related to gh issue #380."""
    n = 1
    eq = desc.examples.get("W7-X")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(3, 3, 3, 6, 6, 6)
    eq.surface = eq.get_surface_at(1.0)

    objective = ObjectiveFunction(
        (
            # just need dummy objective for factorizing constraints
            GenericObjective("0", thing=eq),
        ),
        use_jit=False,
    )
    objective.build()

    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n + 1, :
            ],
        )
    )
    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n + 1, :
    ]
    constraints = (
        FixBoundaryR(eq=eq, modes=R_modes, normalize=False),
        FixBoundaryZ(eq=eq, modes=Z_modes, normalize=False),
        BoundaryRSelfConsistency(eq=eq),
        BoundaryZSelfConsistency(eq=eq),
        FixPressure(eq=eq),
    )
    constraint = ObjectiveFunction(constraints, use_jit=False)
    constraint.build()

    xp, A, b, Z, D, unfixed_idx, project, recover = factorize_linear_constraints(
        objective, constraint
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    np.testing.assert_allclose(x1, x2, atol=atol)
    np.testing.assert_allclose(A @ xp[unfixed_idx], b, atol=atol)
    np.testing.assert_allclose(A @ (x1[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ (x2[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ Z, 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_and_passed_target():
    """Test indexing when passing in specified modes, related to gh issue #380."""
    n = 1
    eq = desc.examples.get("W7-X")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(3, 3, 3, 6, 6, 6)
    eq.surface = eq.get_surface_at(1.0)

    objective = ObjectiveFunction(
        (GenericObjective("0", thing=eq),),
        use_jit=False,
    )
    objective.build()

    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n + 1, :
            ],
        )
    )
    idxs = []
    for mode in R_modes:
        idxs.append(eq.surface.R_basis.get_idx(*mode))
    idxs = np.array(idxs)
    target_R = eq.surface.R_lmn[idxs]

    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n + 1, :
    ]
    idxs = []
    for mode in Z_modes:
        idxs.append(eq.surface.Z_basis.get_idx(*mode))
    idxs = np.array(idxs)
    target_Z = eq.surface.Z_lmn[idxs]

    constraints = (
        FixBoundaryR(eq=eq, modes=R_modes, normalize=False, target=target_R),
        FixBoundaryZ(eq=eq, modes=Z_modes, normalize=False, target=target_Z),
        BoundaryRSelfConsistency(eq=eq),
        BoundaryZSelfConsistency(eq=eq),
        FixPressure(eq=eq),
    )
    constraint = ObjectiveFunction(constraints, use_jit=False)
    constraint.build()

    xp, A, b, Z, D, unfixed_idx, project, recover = factorize_linear_constraints(
        objective, constraint
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    np.testing.assert_allclose(x1, x2, atol=atol)
    np.testing.assert_allclose(A @ xp[unfixed_idx], b, atol=atol)
    np.testing.assert_allclose(A @ (x1[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ (x2[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ Z, 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_axis():
    """Test indexing when passing in specified axis modes, related to gh issue #380."""
    n = 1
    eq = desc.examples.get("W7-X")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(3, 3, 3, 6, 6, 6)
    eq.surface = eq.get_surface_at(1.0)
    eq.axis = eq.get_axis()

    objective = ObjectiveFunction(
        (GenericObjective("0", thing=eq),),
        use_jit=False,
    )
    objective.build()

    R_modes = np.vstack(
        (
            eq.axis.R_basis.modes[np.max(np.abs(eq.axis.R_basis.modes), 1) > n + 1, :],
            [0, 0, 0],
        )
    )
    R_modes = np.flip(R_modes, 0)
    Z_modes = eq.axis.Z_basis.modes[np.max(np.abs(eq.axis.Z_basis.modes), 1) > n + 1, :]
    Z_modes = np.flip(Z_modes, 0)
    constraints = (
        FixAxisR(eq=eq, modes=R_modes, normalize=False),
        FixAxisZ(eq=eq, modes=Z_modes, normalize=False),
        AxisRSelfConsistency(eq=eq),
        AxisZSelfConsistency(eq=eq),
        FixModeR(eq=eq, modes=np.array([[1, 1, 1], [2, 2, 2]]), normalize=False),
        FixModeZ(eq=eq, modes=np.array([[1, 1, -1], [2, 2, -2]]), normalize=False),
        FixModeLambda(eq=eq, modes=np.array([[1, 1, -1], [2, 2, -2]]), normalize=False),
        FixSumModesR(
            eq=eq,
            modes=np.array([[3, 3, 3], [4, 4, 4]]),
            normalize=False,
            sum_weights=np.ones(2),
        ),
        FixSumModesZ(eq=eq, modes=np.array([[3, 3, -3], [4, 4, -4]]), normalize=False),
        FixPressure(eq=eq),
    )
    constraint = ObjectiveFunction(constraints, use_jit=False)
    constraint.build()

    xp, A, b, Z, D, unfixed_idx, project, recover = factorize_linear_constraints(
        objective, constraint
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    np.testing.assert_allclose(x1, x2, atol=atol)
    np.testing.assert_allclose(A @ xp[unfixed_idx], b, atol=atol)
    np.testing.assert_allclose(A @ (x1[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ (x2[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ Z, 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_and_passed_target_axis():
    """Test indexing when passing in specified axis modes, related to gh issue #380."""
    n = 1

    eq = desc.examples.get("W7-X")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(4, 4, 4, 8, 8, 8)
    eq.surface = eq.get_surface_at(1.0)
    eq.axis = eq.get_axis()

    objective = ObjectiveFunction(
        (GenericObjective("0", thing=eq),),
        use_jit=False,
    )
    objective.build()

    R_modes = np.vstack(
        (
            eq.axis.R_basis.modes[np.max(np.abs(eq.axis.R_basis.modes), 1) > n + 1, :],
            [0, 0, 0],
        )
    )
    R_modes = np.flip(R_modes, 0)
    idxs = []
    for mode in R_modes:
        idxs.append(eq.axis.R_basis.get_idx(*mode))
    idxs = np.array(idxs)
    target_R = eq.axis.R_n[idxs]

    Z_modes = eq.axis.Z_basis.modes[np.max(np.abs(eq.axis.Z_basis.modes), 1) > n + 1, :]
    Z_modes = np.flip(Z_modes, 0)
    idxs = []
    for mode in Z_modes:
        idxs.append(eq.axis.Z_basis.get_idx(*mode))
    idxs = np.array(idxs)
    target_Z = eq.axis.Z_n[idxs]

    constraints = (
        AxisRSelfConsistency(eq=eq),
        AxisZSelfConsistency(eq=eq),
        FixAxisR(eq=eq, modes=R_modes, normalize=False, target=target_R),
        FixAxisZ(eq=eq, modes=Z_modes, normalize=False, target=target_Z),
        FixModeR(
            eq=eq,
            modes=np.array([[1, 1, 1], [2, 2, 2]]),
            target=np.array(
                [
                    eq.R_lmn[eq.R_basis.get_idx(*(1, 1, 1))],
                    eq.R_lmn[eq.R_basis.get_idx(*(2, 2, 2))],
                ]
            ),
            normalize=False,
        ),
        FixModeZ(
            eq=eq,
            modes=np.array([[1, 1, -1], [2, 2, -2]]),
            target=np.array(
                [
                    eq.Z_lmn[eq.Z_basis.get_idx(*(1, 1, -1))],
                    eq.Z_lmn[eq.Z_basis.get_idx(*(2, 2, -2))],
                ]
            ),
            normalize=False,
        ),
        FixModeLambda(
            eq=eq,
            modes=np.array([[1, 1, -1], [2, 2, -2]]),
            target=np.array(
                [
                    eq.L_lmn[eq.L_basis.get_idx(*(1, 1, -1))],
                    eq.L_lmn[eq.L_basis.get_idx(*(2, 2, -2))],
                ]
            ),
            normalize=False,
        ),
        FixSumModesR(
            eq=eq,
            modes=np.array([[3, 3, 3], [4, 4, 4]]),
            target=np.array(
                [
                    eq.R_lmn[eq.R_basis.get_idx(*(3, 3, 3))]
                    + eq.R_lmn[eq.R_basis.get_idx(*(4, 4, 4))]
                ]
            ),
            normalize=False,
        ),
        FixSumModesZ(
            eq=eq,
            modes=np.array([[3, 3, -3], [4, 4, -4]]),
            target=np.array(
                [
                    eq.Z_lmn[eq.Z_basis.get_idx(*(3, 3, -3))]
                    + eq.Z_lmn[eq.Z_basis.get_idx(*(4, 4, -4))]
                ]
            ),
            normalize=False,
        ),
        FixPressure(eq=eq),
        FixSumModesLambda(
            eq=eq,
            modes=np.array([[3, 3, -3], [4, 4, -4]]),
            target=np.array(
                [
                    eq.L_lmn[eq.L_basis.get_idx(*(3, 3, -3))]
                    + eq.L_lmn[eq.L_basis.get_idx(*(4, 4, -4))]
                ]
            ),
            normalize=False,
        ),
    )
    constraint = ObjectiveFunction(constraints, use_jit=False)
    constraint.build()

    xp, A, b, Z, D, unfixed_idx, project, recover = factorize_linear_constraints(
        objective, constraint
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    np.testing.assert_allclose(x1, x2, atol=atol)
    np.testing.assert_allclose(A @ xp[unfixed_idx], b, atol=atol)
    np.testing.assert_allclose(A @ (x1[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ (x2[unfixed_idx] / D[unfixed_idx]), b, atol=atol)
    np.testing.assert_allclose(A @ Z, 0, atol=atol)


@pytest.mark.unit
def test_FixBoundary_with_single_weight():
    """Test Fixing boundary with only a single, passed weight."""
    eq = Equilibrium()
    w = 1.1
    FixZ = FixBoundaryZ(eq=eq, modes=np.array([[0, -1, 0]]), weight=w)
    FixZ.build()
    np.testing.assert_array_equal(FixZ.weight.size, 1)
    np.testing.assert_array_equal(FixZ.weight, w)
    FixR = FixBoundaryR(eq=eq, modes=np.array([[0, 1, 0]]), weight=w)
    FixR.build()
    np.testing.assert_array_equal(FixR.weight.size, 1)
    np.testing.assert_array_equal(FixR.weight, w)


@pytest.mark.unit
def test_FixBoundary_passed_target_no_passed_modes_error():
    """Test Fixing boundary with no passed-in modes."""
    eq = Equilibrium()
    FixZ = FixBoundaryZ(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixZ.build()
    FixZ = FixBoundaryZ(eq=eq, modes=False, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixZ.build()
    FixR = FixBoundaryR(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixR.build()
    FixR = FixBoundaryR(eq=eq, modes=False, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixR.build()


@pytest.mark.unit
def test_FixAxis_passed_target_no_passed_modes_error():
    """Test Fixing Axis with no passed-in modes."""
    eq = Equilibrium()
    FixR = FixAxisR(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixR.build()
    FixR = FixAxisR(eq=eq, modes=False, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixR.build()
    FixZ = FixAxisZ(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixZ.build()
    FixZ = FixAxisZ(eq=eq, modes=False, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixZ.build()


@pytest.mark.unit
def test_FixMode_passed_target_no_passed_modes_error():
    """Test Fixing Modes with no passed-in modes."""
    eq = Equilibrium()
    FixZ = FixModeZ(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixZ.build()
    FixR = FixModeR(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixR.build(eq)
    FixL = FixModeLambda(eq=eq, modes=True, target=np.array([0, 0]))
    with pytest.raises(ValueError):
        FixL.build(eq)


@pytest.mark.unit
def test_FixSumModes_passed_target_too_long():
    """Test Fixing Modes with more than a size-1 target."""
    # TODO (#1399): remove this test if FixSumModes is generalized
    # to accept multiple targets and sets of modes at a time
    eq = Equilibrium(L=3, M=4)
    with pytest.raises(ValueError):
        FixSumModesZ(
            eq, modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]])
        )
    with pytest.raises(ValueError):
        FixSumModesR(
            eq, modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]])
        )
    with pytest.raises(ValueError):
        FixSumModesLambda(
            eq, modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]])
        )


@pytest.mark.unit
def test_FixSumModes_False_or_None_modes():
    """Test Fixing Sum Modes without specifying modes or All modes."""
    eq = Equilibrium(L=3, M=4)
    with pytest.raises(ValueError):
        FixSumModesR(eq, modes=False)
    with pytest.raises(ValueError):
        FixSumModesR(eq, modes=None)
    with pytest.raises(ValueError):
        FixSumModesZ(eq, modes=False)
    with pytest.raises(ValueError):
        FixSumModesZ(eq, modes=None)
    with pytest.raises(ValueError):
        FixSumModesLambda(eq, modes=False)
    with pytest.raises(ValueError):
        FixSumModesLambda(eq, modes=None)


def _is_any_instance(things, cls):
    return any([isinstance(t, cls) for t in things])


@pytest.mark.unit
def test_FixAxis_util_correct_objectives():
    """Test util for fix axis constraints."""
    eq = Equilibrium()
    cs = get_fixed_axis_constraints(eq)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixPressure)
    assert _is_any_instance(cs, FixCurrent)

    eq = Equilibrium(electron_temperature=1, electron_density=1, iota=1)
    cs = get_fixed_axis_constraints(eq)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixElectronDensity)
    assert _is_any_instance(cs, FixElectronTemperature)
    assert _is_any_instance(cs, FixIonTemperature)
    assert _is_any_instance(cs, FixAtomicNumber)
    assert _is_any_instance(cs, FixIota)


@pytest.mark.unit
def test_FixNAE_util_correct_objectives():
    """Test util for fix NAE constraints."""
    eq = Equilibrium()
    qsc = Qsc.from_paper("precise QA")
    cs = get_NAE_constraints(eq, qsc)
    for c in cs:
        c.build()
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixNearAxisR)
    assert _is_any_instance(cs, FixNearAxisZ)
    assert _is_any_instance(cs, FixPressure)
    assert _is_any_instance(cs, FixCurrent)

    cs = get_NAE_constraints(eq, qsc, fix_lambda=1)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixNearAxisR)
    assert _is_any_instance(cs, FixNearAxisZ)
    assert _is_any_instance(cs, FixNearAxisLambda)
    assert _is_any_instance(cs, FixPressure)
    assert _is_any_instance(cs, FixCurrent)

    eq = Equilibrium(electron_temperature=1, electron_density=1, iota=1)
    cs = get_NAE_constraints(eq, qsc)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixNearAxisR)
    assert _is_any_instance(cs, FixNearAxisZ)
    assert _is_any_instance(cs, FixElectronDensity)
    assert _is_any_instance(cs, FixElectronTemperature)
    assert _is_any_instance(cs, FixIonTemperature)
    assert _is_any_instance(cs, FixAtomicNumber)
    assert _is_any_instance(cs, FixIota)


@pytest.mark.unit
def test_fix_omni_indices():
    """Test that omnigenity parameters are constrained properly.

    Test for GH issue #768.
    """
    NFP = 3
    field = OmnigenousField(
        L_B=1, M_B=5, L_x=2, M_x=3, N_x=4, NFP=NFP, helicity=(1, NFP)
    )

    # no indices
    constraint = FixOmniWell(field=field, indices=False)
    constraint.build()
    assert constraint.dim_f == 0
    constraint = FixOmniMap(field=field, indices=False)
    constraint.build()
    assert constraint.dim_f == 0

    # all indices
    constraint = FixOmniWell(field=field, indices=True)
    constraint.build()
    assert constraint.dim_f == field.B_lm.size
    constraint = FixOmniMap(field=field, indices=True)
    constraint.build()
    assert constraint.dim_f == field.x_lmn.size

    # specified indices
    indices = np.arange(3, 8)
    constraint = FixOmniWell(field=field, indices=indices)
    constraint.build()
    assert constraint.dim_f == indices.size
    constraint = FixOmniMap(field=field, indices=indices)
    constraint.build()
    assert constraint.dim_f == indices.size


@pytest.mark.unit
def test_fix_omni_Bmax():
    """Test that omnigenity parameters are constrained for B_max to be a straight line.

    Test for GH issue #1266.
    """

    def _test(M_x, N_x, NFP, sum):
        field = OmnigenousField(L_x=2, M_x=M_x, N_x=N_x, NFP=NFP, helicity=(1, NFP))
        constraint = FixOmniBmax(field=field)
        constraint.build()
        assert constraint.dim_f == (2 * field.N_x + 1) * (field.L_x + 1)
        # 0 - 2 + 4 - 6 + 8 ...
        np.testing.assert_allclose(constraint._A @ field.x_basis.modes[:, 1], sum)

    _test(M_x=6, N_x=3, NFP=1, sum=-4)
    _test(M_x=9, N_x=4, NFP=2, sum=4)
    _test(M_x=12, N_x=5, NFP=3, sum=6)


@pytest.mark.unit
def test_fix_parameters_input_order(DummyStellarator):
    """Test that FixParameters preserves the input indices and target ordering."""
    eq = load(load_from=str(DummyStellarator["output_path"]), file_format="hdf5")
    default_target = eq.Rb_lmn

    # default objective
    obj = FixBoundaryR(eq)
    obj.build()
    np.testing.assert_allclose(obj.target, default_target)

    # manually specify default
    obj = FixBoundaryR(eq, modes=eq.surface.R_basis.modes)
    obj.build()
    np.testing.assert_allclose(obj.target, default_target)

    # reverse order
    obj = FixBoundaryR(eq, modes=np.flipud(eq.surface.R_basis.modes))
    obj.build()
    np.testing.assert_allclose(obj.target, np.flipud(default_target))

    # custom order
    obj = ObjectiveFunction(
        FixBoundaryR(eq, modes=np.array([[0, 0, 0], [0, 1, 0], [0, 1, 1]]))
    )
    obj.build()
    np.testing.assert_allclose(obj.target_scaled, np.array([3, 1, 0.3]))
    np.testing.assert_allclose(obj.compute_scaled_error(obj.x(eq)), np.zeros(obj.dim_f))

    # custom target
    obj = ObjectiveFunction(
        FixBoundaryR(
            eq,
            modes=np.array([[0, 0, 0], [0, 1, 0], [0, 1, 1]]),
            target=np.array([0, -1, 0.5]),
        )
    )
    obj.build()
    np.testing.assert_allclose(
        obj.compute_scaled_error(obj.x(eq)), np.array([3, 2, -0.2])
    )


@pytest.mark.unit
def test_fix_subset_of_params_in_collection(DummyMixedCoilSet):
    """Tests FixParameters fixing a subset of things in the collection."""
    coilset = load(load_from=str(DummyMixedCoilSet["output_path"]), file_format="hdf5")

    params = [
        [
            {"current": True},
        ],
        {"shift": True, "rotmat": True},
        {"X_n": np.array([1, 2]), "Y_n": False, "Z_n": np.array([0])},
        {},
    ]
    target = np.concatenate(
        (
            np.array([3]),
            np.eye(3).flatten(),
            np.array([0, 0, 0]),
            np.eye(3).flatten(),
            np.array([0, 0, 1]),
            np.eye(3).flatten(),
            np.array([0, 0, 2]),
            np.array([10, 2, -2]),
        )
    )

    obj = FixParameters(coilset, params)
    obj.build()
    np.testing.assert_allclose(obj.target, target)


@pytest.mark.unit
def test_fix_coil_current(DummyMixedCoilSet):
    """Tests FixCoilCurrent."""
    coilset = load(load_from=str(DummyMixedCoilSet["output_path"]), file_format="hdf5")

    # fix a single coil current
    obj = FixCoilCurrent(coil=coilset.coils[1].coils[0])
    obj.build()
    assert obj.dim_f == 1
    np.testing.assert_allclose(obj.target, -1)

    # fix all coil currents
    obj = FixCoilCurrent(coil=coilset)
    obj.build()
    assert obj.dim_f == 6
    np.testing.assert_allclose(obj.target, [3, -1, -1, -1, 2, 1])

    # only fix currents of some coils in the coil set
    obj = FixCoilCurrent(
        coil=coilset, indices=[[True], [True, False, True], True, False]
    )
    obj.build()
    assert obj.dim_f == 4
    np.testing.assert_allclose(obj.target, [3, -1, -1, 2])


@pytest.mark.unit
def test_fix_sum_coil_current(DummyMixedCoilSet):
    """Tests FixSumCoilCurrent."""
    coilset = load(load_from=str(DummyMixedCoilSet["output_path"]), file_format="hdf5")

    # sum a single coil current
    obj = FixSumCoilCurrent(coil=coilset.coils[1].coils[0])
    obj.build()
    params = coilset.coils[1].coils[0].params_dict
    np.testing.assert_allclose(obj.compute(params), -1)

    # sum all coil currents
    obj = FixSumCoilCurrent(coil=coilset)
    obj.build()
    params = coilset.params_dict
    np.testing.assert_allclose(obj.compute(params), 3)
    # the default target should be the original sum
    np.testing.assert_allclose(obj.compute_scaled_error(params), 0)

    # only sum currents of some coils in the coil set
    obj = FixSumCoilCurrent(
        coil=coilset, indices=[[True], [True, False, True], True, False]
    )
    obj.build()
    np.testing.assert_allclose(obj.compute(params), 3)


@pytest.mark.unit
def test_linear_objective_from_user_on_collection(DummyCoilSet):
    """Test LinearObjectiveFromUser on an OptimizableCollection."""
    # test that LinearObjectiveFromUser can be used for the same functionality as
    # FixSumCoilCurrent to sum all currents in a CoilSet

    coilset = load(load_from=str(DummyCoilSet["output_path_asym"]), file_format="hdf5")
    params = coilset.params_dict

    obj1 = FixSumCoilCurrent(coil=coilset)
    obj1.build()

    obj2 = LinearObjectiveFromUser(
        lambda params: jnp.sum(jnp.array([param["current"] for param in params])),
        coilset,
    )
    obj2.build()

    np.testing.assert_allclose(obj1.compute(params), obj2.compute(params))
